from .lazyClassifier import LazyClassifier
from .lazyRegressor import LazyRegressor
from .lazyMTS import LazyMTS
from .lazydeepClassifier import LazyDeepClassifier
from .lazydeepRegressor import LazyDeepRegressor
from .lazydeepMTS import LazyDeepMTS

__all__ = [
    "LazyClassifier",
    "LazyRegressor",
    "LazyMTS",
    "LazyDeepClassifier",
    "LazyDeepRegressor",
    "LazyDeepMTS",
]
