from .ridge2Classifier import Ridge2Classifier
from .ridge2Regressor import Ridge2Regressor
from .ridge2MultitaskClassifier import Ridge2MultitaskClassifier

__all__ = ["Ridge2Classifier", "Ridge2Regressor", "Ridge2MultitaskClassifier"]
